Needs["xAct`xTensor`"];


(****** extract the principal part from a set of equations and return it as a vector ******)
(* highestDerivs: list of the highest derivative orders of the fields
   f:             list of fields
   eqn:           list of equations
*)
extractPP::"variableNoxTensor" = "fields must be xTensor's.";
extractPP::"otherHighestDerivsDetected" = "Warning: Detected other highest derivatives in the equations than specified: `1` instead of `2`.";
Options[extractPP] := {ReshapeFunction -> Hold[ChristoffelToGradMetric[ChangeCovD[ChangeCurvature[#]]]&]}
extractPP[highestDerivs_List][{f__?xTensorQ}][eqn_List] :=
Module[{reshapeFunction = ReshapeFunction /. Options[extractPP],
        rhsEquations,
        actualHighestDerivs,
        inverseEquationOrders,
        maxHighestDerivs = highestDerivs},

       (********** all fields must be xTensors ***********)
       If[Not @ (And @@ (xTensorQ[#]& /@ {f})),
          Message[extractPP::"variableNoxTensor"];Throw[Null]];

       (********** calculating the right hand sides of the equations *******)
       rhsEquations = ReleaseHold[reshapeFunction /@ eqn];
       (*Print[rhsEquations];*)
       Print["** equations derived"];

       (********* the highest derivatives of each variable in each equation ***********)
       actualHighestDerivs = findHighestOrderDerivOn[{f}][#]& /@ rhsEquations;
       If[(Max /@ Transpose[actualHighestDerivs]) != maxHighestDerivs,
          Message[extractPP::"otherHighestDerivsDetected",
                  (Max /@ Transpose[actualHighestDerivs]),
                  maxHighestDerivs]];

       (**** if an equation contains only lower order derivatives
             then we return those leading terms ****)
       inverseEquationOrders = (maxHighestDerivs -
                                Min[maxHighestDerivs - #])& /@ actualHighestDerivs;

       (******* drop lower order terms ***********)
       Return[DropLowerOrdersThan[inverseEquationOrders[[#]]][{f}][rhsEquations[[#]]]& /@
              Range[Length[rhsEquations]]];
      ];



(********* calculating lhs and rhs from a Hamiltonian **********)
(* H:               Hamiltonian
   positions:       list of positions variables
   momenta:         list of momentum variables
   extFields:       external fields (not yet implemented)
 example: calcHamLhsRhs[HS[], {metrich}, {pi}, {}]
*)
calcHamLhsRhs::"unequalSize" = "The number of positions and momenta must be the same, we got `1` and `2`.";
calcHamLhsRhs[H_, positions_List, momenta_List, extFields_List, freeInds_List,
              options___] :=
Module[{fieldIndices =  GetStandardIndicesOfTensor[#,{}, freeInds]& /@ positions,
        indexedCanonicalFields = Join[momenta, positions],
        rhsSigns = Join[Table[ 1, {Length[momenta]}],
                        Table[-1, {Length[positions]}]],
        lhs,
        rhs},
       If[Length[positions] != Length[momenta],
           Message[calcHamLhsRhs::"unequalSize", positions, momenta];
           Throw[Null]
         ];
       fieldIndices = Join[fieldIndices, -fieldIndices];
       indexedCanonicalFields = indexedCanonicalFields[[#]]@@(-fieldIndices)[[#]]& /@
                                   Range[Length[fieldIndices]];
       Validate /@ indexedCanonicalFields;
       rhs = VarD[#][H]& /@ indexedCanonicalFields;
       rhs = (rhsSigns[[#]] rhs[[#]])& /@ Range[Length[rhs]];
       lhs = ParamD[t] /@ ((Join[positions,momenta][[#]]@@fieldIndices[[#]])& /@
                           Range[Length[fieldIndices]]);
       {lhs,rhs}
      ];
calcHamLhsRhs[H_, positions_List, momenta_List, extFields_List] :=
calcHamLhsRhs[H, positions, momenta, extFields, {}];



(********* the principal part of right hand sides from a Hamiltonian **********)
(* H:               Hamiltonian
   highestPosDeriv: list of highest order derivatives to position variables
   positions:       list of positions variables
   highestMomDeriv: list of highest order derivatives to momentum variables
   momenta:         list of momentum variables
   highestDerivsEx: highest derivatives on external fields (not yet implemented)
   extFields:       external fields (not yet implemented)
 example: findHamPPRhs[HS[], {2}, {metrich}, {1}, {pi}, {}, {}]
*)
calcHamRhsPP[H_, highestPosDeriv_List, positions_List,
                 highestMomDeriv_List, momenta_List,
                 highestExtDeriv_List, extFields_List,
             freeInds_List,
             options___] :=
Module[{rhs},
       rhs = calcHamLhsRhs[H, positions, momenta, extFields,
                           freeInds,
                           options][[2]];
       extractPP[Join[highestPosDeriv, highestMomDeriv]][Join[positions, momenta]][rhs]
      ];
calcHamRhsPP[H_, highestPosDeriv_List, positions_List,
                 highestMomDeriv_List, momenta_List,
                 highestExtDeriv_List, extFields_List] :=
calcHamRhsPP[H, highestPosDeriv, positions,
                highestMomDeriv, momenta,
                highestExtDeriv, extFields,
             {}];

(****** calculation of M, V, F, FT ******)
calcHamRhsPPMatrixBlocks[H_, highestPosDeriv_List, positions_List,
                             highestMomDeriv_List, momenta_List,
                             highestExtDeriv_List, extFields_List,
                         derivInds_List, rowInds_List, colInds_List,
                         options___] :=
Module[{rhsPP, nthOrderFields, eqnOrders, nthOrderFieldsPositions, nthOrderEqnPositions,
        PPM, Minv, V, F, FT, counter
        },
       rhsPP = calcHamRhsPP[H, highestPosDeriv, positions,
                               highestMomDeriv, momenta,
                               highestExtDeriv, extFields,
                            rowInds,
                            options];
       nthOrderFields = Table[0,{Max[Join[highestPosDeriv,highestMomDeriv]]}];
       For[counter=Min[Join[highestPosDeriv,highestMomDeriv]],
           counter<=Max[Join[highestPosDeriv,highestMomDeriv]],
           counter++,
           nthOrderFields[[counter]] =
           (#[[1]])& /@ Select[Thread[List[Range[Length[Join[positions,momenta]]],
                                           Join[highestPosDeriv,highestMomDeriv]]],
                               (#[[2]]==counter)&]
          ];
       nthOrderFieldsPositions = Position[Join[highestPosDeriv,highestMomDeriv], #] & /@
                                 Union[Join[highestPosDeriv,highestMomDeriv]];
       nthOrderFieldsPositions = Flatten /@ nthOrderFieldsPositions;

       eqnOrders = (Max /@ ((findHighestOrderDerivOn[Join[positions,momenta]][#] -
                                Join[highestPosDeriv,highestMomDeriv])& /@ rhsPP)) +
                      Max[Join[highestPosDeriv,highestMomDeriv]];
       nthOrderEqnPositions = Position[eqnOrders, #] & /@
                              Union[eqnOrders];
       nthOrderEqnPositions = Flatten /@ nthOrderEqnPositions;

       PPM = PPMatrix[Join[highestPosDeriv,highestMomDeriv],
                      Join[positions,momenta],
                      derivInds, colInds][rhsPP];

       Minv = Part[PPM, nthOrderEqnPositions[[1]], nthOrderFieldsPositions[[1]]];
       V = Part[PPM, nthOrderEqnPositions[[2]], nthOrderFieldsPositions[[2]]];
       F = Part[PPM, nthOrderEqnPositions[[1]], nthOrderFieldsPositions[[2]]];
       FT = Part[PPM, nthOrderEqnPositions[[2]], nthOrderFieldsPositions[[1]]];
       {Minv, V, F, FT}
      ];



(*********************************************************************************)
(* PPMatrix: derive the principal part matrix of given principal part rhs *)
PPMatrix::"PPVecWithPD" = "Warning: the principal part still contains derivatives."
PPMatrix::"sameIndices" = "Error: the sets of row, column and PD indices must not intersect, but there is a non empty intersection of `1`, `2` and `3`."
PPMatrix::"Reorder" = "Warning: probably reordering indices, before: `1`, after: `2`.";
PPMatrix[highestDerivs_List, {fields__?xTensorQ}, {PDIndices___?AIndexQ}, {colIndices___?AIndexQ}][eqn_List] :=
Module[{uniquePDIndices = Union[{PDIndices}], (** list of PD indices **)
        maxDerivOrder, (** highest order derivative in all equations **)
        uniqueColIndices = Union[{colIndices}], (** list of column indices **)
        maxRank, (** rank of the highest rank tensor in {fields} **)
        rowIndices, (** list of row indices **)
        eqnRowIndices, (** list of row indices for each equation **)
        fieldColIndices, (** list of column for each field **)
        PPDerivOrderMatrix, (** matrix of derivative orders in the principal part **)
        dummyRow, (** if some equation does not contain any of the fields then we must replace the negative row **)
        PPPDIndicesMatrix, (** matrix of derivative indices **)
        unitedDerivOrderMatrix, (** matrix of derivative orders for definition of new tensors **)
        unitedPDIndicesMatrix, (** matrix of derivative indices for definition of new tensors **)
        eqnReplaceRuleNumbers, (** list that describes in which equation which rule must be applied **)
        origTensorNamesMatrix, (** matrix of names of fields **)
        origTensorIndicesMatrix, (** matrix of column indices of fields **)
        origIndexedTensorsMatrix, (** matrix fields applied to standard indices**)
        newTensorNamesMatrix, (** matrix of names of new dummy tensors **)
        newTensorIndicesMatrix, (** matrix of indices of new dummy tensors **)
        newIndexedTensorsMatrix, (** matrix of names applied to indices of new dummy tensors **)
        newTensorSymmetriesMatrix, (** matrix of symmetries of new dummy tensors **)
        new, (** to indicate the dummy tensors **)
        del, (** to deal with the problem of IndexCoefficient with traces **)
        delIndices, (** two indices for the definition of del **)
        dummyRulesLhs, (** left hand side of replacement rules **)
        dummyRules, (** replacement rules **)
        eqnDummyRules, (** list of replacement rules that corresponds to list of equations **)
        replacedEqn, (** equations with replaced highest order derivatives **)
        counter, (** counter for loops **)
        row, col, (** to build the final matrix **)
        PP (** the result **)
       },

       (*** check ordering and signs of indices ***)
       If[Sort[RemoveSign[{PDIndices}]] =!= RemoveSign[{PDIndices}],
          Message[PPMatrix::"Reorder", RemoveSign[{PDIndices}], Sort[RemoveSign[{PDIndices}]]];
         ];
       If[Sort[RemoveSign[{colIndices}]] =!= RemoveSign[{colIndices}],
          Message[PPMatrix::"Reorder", RemoveSign[{colIndices}], Sort[RemoveSign[{colIndices}]]];
         ];

       (*** the row indices ***)
       eqnRowIndices = (List@@IndicesOf[Free][#])& /@ eqn;
       rowIndices = Union[RemoveSign[Flatten[eqnRowIndices]]];

       (*** are row, column and PD indices different? ***)
       If[Length[Union[uniquePDIndices, uniqueColIndices, rowIndices]]<
          (Length[uniquePDIndices] + Length[uniqueColIndices] + Length[rowIndices]),
          Message[PPMatrix::"sameIndices", rowIndices, {colIndices}, {PDIndices}];
          Throw[Null];
         ];

       (*** how many column indices do we need and which? ***)
       maxRank = Max[Length /@ (SlotsOfTensor /@ {fields})];
       If[maxRank > Length[uniqueColIndices],
          uniqueColIndices = Join[uniqueColIndices,
                           GetIndicesOfVBundle[
                              TangentBundleOfManifold @
                                       Intersection[Sequence @@ (DependenciesOfTensor /@ {fields}),
                                                    $Manifolds][[1]],
                              maxRank-Length[uniqueColIndices],
                              Join[uniqueColIndices, uniquePDIndices, rowIndices]]]
         ];
       uniqueColIndices = Sort[uniqueColIndices];
       fieldColIndices = GetStandardIndicesOfTensor[#, {}, uniqueColIndices]& /@ {fields};

       (*** how many derivative indices do we need and which? ***)
       maxDerivOrder = Max[findHighestOrderDerivOn[{fields}][eqn]];
       If[maxDerivOrder > Length[uniquePDIndices],
          uniquePDIndices = Join[uniquePDIndices,
                          GetIndicesOfVBundle[
                              TangentBundleOfManifold @
                                       Intersection[Sequence @@ (DependenciesOfTensor /@ {fields}),
                                                    $Manifolds][[1]],
                              maxDerivOrder-Length[uniquePDIndices],
                              Join[uniqueColIndices, uniquePDIndices, rowIndices]]]
         ];
       delIndices = uniquePDIndices;
       If[Length[delIndices] < 2,
          delIndices = Join[delIndices,
                          GetIndicesOfVBundle[
                              TangentBundleOfManifold @
                                       Intersection[Sequence @@ (DependenciesOfTensor /@ {fields}),
                                                    $Manifolds][[1]],
                              2-Length[delIndices],
                              delIndices]],
          delIndices = delIndices[[1;;2]];
         ];

       (*** in this matrix the orders of the derivatives in the PP is stored ***)
       PPDerivOrderMatrix = Outer[Plus,
                                Max /@ ((findHighestOrderDerivOn[{fields}][#] - highestDerivs)& /@ eqn),
                                highestDerivs];
       (*** if some equation does not contain any of the fields then we must replace the negative row ***)
       If[Max[Flatten[PPDerivOrderMatrix]] < 0,
          Return[Table[0, {Length[eqn]}, {Length[{fields}]}]]];
       dummyRow = {-1};
       counter = 1;
       While[Max[dummyRow]<0,
             dummyRow = PPDerivOrderMatrix[[counter]];
             counter++];
       For[counter=1, counter<=Length[PPDerivOrderMatrix], counter++,
           If[Max[PPDerivOrderMatrix[[counter]]]<0,
              PPDerivOrderMatrix[[counter]] = dummyRow]
          ];

       (*** construct matrix of derivative indices in the PP ***)
       PPPDIndicesMatrix = Map[Take[-uniquePDIndices, {1, #}]&, PPDerivOrderMatrix, {2}];





       (****** construction of replacement rules ******)

       (*** unite equal derivative orders from different equations ***)
       unitedDerivOrderMatrix = Union[PPDerivOrderMatrix];

       (*** which replacement rule will we need to extract the derivatives in the equations? ***)
       eqnReplaceRuleNumbers = (Sequence @@@ Position[PPDerivOrderMatrix, #])& /@ unitedDerivOrderMatrix;
       eqnReplaceRuleNumbers = Position[eqnReplaceRuleNumbers, #][[1,1]]& /@ Range[Length[eqn]];

       (*** construct matrix of derivative indices for definition of new tensors ***)
       unitedPDIndicesMatrix = Map[Take[-uniquePDIndices, {1, #}]&, unitedDerivOrderMatrix, {2}];



       (*** matrices of indices and tensor names without derivatives ***)
       origTensorNamesMatrix = Table[{fields}, {Length[unitedDerivOrderMatrix]}];
       origTensorIndicesMatrix = Table[fieldColIndices, {Length[unitedDerivOrderMatrix]}];
       origIndexedTensorsMatrix = MapThread[Apply[#1, #2]&,
                                            {origTensorNamesMatrix, origTensorIndicesMatrix}, 2];

       (*** matrix of new tensor names ***)
       newTensorNamesMatrix = MapThread[GiveSymbol[#1, #2, new]&,
                                        {Map[StringJoin @@ Table["d", {#}]&, unitedDerivOrderMatrix, {2}],
                                         Table[{fields}, {Length[unitedDerivOrderMatrix]}]},2];

       (*** left hand side of replace rules ***)
       dummyRulesLhs = MapThread[If[Length[#1]>0, (Composition @@ Map[PD, #1])[#2], #2]&,
                                 {unitedPDIndicesMatrix,origIndexedTensorsMatrix}, 2];

       (*** symmetries of the new tensors ***)
       newTensorSymmetriesMatrix = Map[SymmetryOf, dummyRulesLhs, {2}];
       (*** matrix of new tensor indices ***)
       newTensorIndicesMatrix =
       MapThread[(Table[xAct`xTensor`Private`slot[counter], {counter, #1+Length[#2]}] /. #3[[3]])&,
                 {unitedDerivOrderMatrix, origTensorIndicesMatrix, newTensorSymmetriesMatrix}, 2];

       (*** apply tensor names to indices ***)
       newIndexedTensorsMatrix = MapThread[Apply[#1, #2]&,
                                           {newTensorNamesMatrix, newTensorIndicesMatrix}, 2];

       (*** defining new dummy tensors ***)
       MapThread[DefTensor[#1,#2,#3[[4]]]&,
                 {newIndexedTensorsMatrix,
                  Table[DependenciesOfTensor /@ {fields}, {Length[newTensorNamesMatrix]}],
                  newTensorSymmetriesMatrix}, 2];

       (*** constructing replacement rules ***)
       dummyRules = MapThread[MakeRule[{#1, #2}, MetricOn->None]&,
                              {dummyRulesLhs, newIndexedTensorsMatrix}, 2];
       dummyRules = Apply[Flatten, dummyRules, {2}];
       (*** list of replace rules for equations ***)
       eqnDummyRules = Table[dummyRules[[eqnReplaceRuleNumbers[[i]]]], {i, Length[eqn]}];

       (*** equations with replaced derivatives ***)
       replacedEqn = MapThread[ReplaceAll, {eqn, eqnDummyRules}, 1];

       If[Not@FreeQ[replacedEqn,PD],
          Message[PPMatrix::"PPVecWithPD"];
          Print[replacedEqn]];

       (*** deal with the problem of IndexCoefficient with traces ***)
       DefTensor[del[delIndices[[1]], -delIndices[[2]]],
                 Union[Join @@ (DependenciesOfTensor /@ $Tensors)]];
       replacedEqn = ReplaceRepeated[replacedEqn,
                                     T_?xTensorQ[a___,b_,c___,-b_,d___] :> T[a,i,c,-b,d] del[b, -i]];
       replacedEqn = ReplaceRepeated[replacedEqn,
                                     T_?xTensorQ[a___,-b_,c___,b_,d___] :> T[a,-b,c,i,d] del[b, -i]];

       (*** construct result matrix ***)
       PP = Table[IndexCoefficient[replacedEqn[[row]],
                                   newIndexedTensorsMatrix[[eqnReplaceRuleNumbers[[row]], col]]],
                  {row, Length[eqn]}, {col, Length[{fields}]}];
       PP = PP /. del->delta;

       (*** clean ***)
       UndefTensor[del];
       UndefTensor /@ Flatten[newTensorNamesMatrix];

       PP
];
PPMatrix[highestDerivs_List, {fields__?xTensorQ}, {PDIndices___?AIndexQ}][eqn_List] :=
PPMatrix[highestDerivs, {fields}, {PDIndices}, {}][eqn];
PPMatrix[highestDerivs_List, {fields__?xTensorQ}][eqn_List] :=
PPMatrix[highestDerivs, {fields}, {}, {}][eqn];












(********** find the number of partial derivatives on a field **********)
Options[PDQ] = {Patterns -> {PD[-(a_Symbol)?AIndexQ],
                             ParamD[(t__Symbol)?ParameterQ]}};
PDQ[deriv_, options___] :=
Module[{patterns = (Patterns /. {options}) /. Options[PDQ]},
       If[Or @@ (MatchQ[deriv, #] & /@ patterns),
          True,
          False]
      ];

findHighestOrderDerivOn[f_?xTensorQ][term_, options___] :=
Block[{list = List @@ term},
      If[FreeQ[term,f], Return[-1]];
      If[(Head @ term) === List, Return[((findHighestOrderDerivOn[f][#])& /@ term)]];
      (*If[list === term, Return[0]]; (**** if term is NULL ****)*)
      If[MatchQ[term,HoldPattern[(PD_?PDQ)[A_?(Not@FreeQ[#,f]&)]]],
         (Length[Head[term]] + findHighestOrderDerivOn[f][term[[1]]]),
         Max @@ {0, Sequence @@ ((findHighestOrderDerivOn[f][#])& /@ list)}
        ]
     ];
findHighestOrderDerivOn[f_List][term_] := (findHighestOrderDerivOn[#][term])& /@ f;


(********** extracting expressions with higher derivatives **********)
DropLowerOrdersThan[MinimalDerivOrders_List][{f__?xTensorQ}][sum_Plus] :=
DropLowerOrdersThan[MinimalDerivOrders][{f}] /@ sum;
DropLowerOrdersThan[MinimalDerivOrders_List][{f__?xTensorQ}][prod_Times] :=
Block[{list = List @@ prod,
       minOrderList},
      minOrderList = Min[MinimalDerivOrders-findHighestOrderDerivOn[{f}][#]]& /@ list;
      Return[Times@@If[Min[minOrderList] > 0,
                       0,
                       If[minOrderList[[#]]>0,
                          list[[#]],
                          DropLowerOrdersThan[MinimalDerivOrders][{f}][list[[#]]]]& /@
                        Range[Length[list]]]];
     ];
DropLowerOrdersThan[MinimalDerivOrders_List][{f__?xTensorQ}][term_] :=
If[Min@(MinimalDerivOrders-findHighestOrderDerivOn[{f}][term])>0,0,term];



(********* auxiliary functions *********)

GetStandardIndicesOfTensor::"Reorder" = "Warning: probably reordering indices, before: `1`, after: `2`.";
GetStandardIndicesOfTensor[T_?xTensorQ, exclude_List, include_List] :=
Module[{slots, signs, indices, counter,
        excludeInds = Union[RemoveSign[exclude]],
        includeInds = Union[Complement[RemoveSign[include], RemoveSign[exclude]]],
        thisSlotIncludeInds},
       If[Sort[RemoveSign[include]] =!= RemoveSign[include],
          Message[GetStandardIndicesOfTensor::"Reorder", RemoveSign[include], Sort[RemoveSign[include]]];
         ];
       slots = SlotsOfTensor[T];
       signs = If[Head[#]===Times,-1,1]& /@ slots;
       indices = {};
       For[counter = 1, counter <= Length[slots], counter++,
           thisSlotIncludeInds =
           Intersection[Flatten[IndicesOfVBundle[slots[[counter]]]],
                        includeInds];
           If[Length[Complement[thisSlotIncludeInds, indices]]>0,
              indices = Join[indices, {Complement[thisSlotIncludeInds, indices][[1]]}],
              indices = Join[indices,
                             GetIndicesOfVBundle[slots[[counter]], 1, Join[indices,excludeInds]]
                            ]
             ]
          ];
       indices*signs];
GetStandardIndicesOfTensor[T_?xTensorQ, exclude_List] :=
GetStandardIndicesOfTensor[T, exclude, {}];

RemoveSign[indexList_] := If[Head[#]===Times,-#,#]& /@ indexList;
